function afemReduced(example,alpha,Q,compare,theta,minNrDoF)
% Function for running benchmarks for reduced mixed method 
% Input:   example    problem to solve, for options see below
%          alpha      real parameter in reduced formulation
%          Q          local projection in reduced formulation, for options see below
%          compare    compare to equivalent dPG (compare = 1) or not ( = 0)
%          theta      bulk parameter, uniform refinement if set to 1
%          minNrDoF   number of degrees of freedom to be reached minimally
    
    % Add current path
    addpath(genpath(pwd));

    %% Set data if none is given
    if( nargin == 0)
        % Options for problem
%         example = 'LshapeDirac';        
%         example = 'LshapeDiracReverse';        
%         example = 'LshapeExact';        
        example = 'OneRHS';        
%         example = 'SlitExact';        
%         example = 'SquareExact';        
%         example = 'SquareExact2';        
%         example = 'SquareExact3';        
%         example = 'Waterfall';        

        compare = 0;
        
        alpha = 1;
        Q = 'identity'; % Q = identity mapping
%         Q = 'piZero'; % Q = L^2 projection onto piecewise constants

        % Set Parameters
        theta = 1;
        minNrDoF = 10000; 
    end

    %% Collect data for example
    data = eval(example);
    c4n = data{1};
    n4e = data{2};
    n4sDb = data{3};
    n4sNb = data{4};
    f = data{5};
    u4Db = data{6};
    g = data{7};
    exSolKnown = data{8};
    % optional exact solution data
    if exSolKnown == 1
        uExact = data{9};
        graduExact = data{10}; 
    end 
    
    %% Output of data
    % Set name for folder in which to save data
    if theta == 1
        folder = ['output/reduced_',example,'_Q',Q,'_alpha',num2str(alpha),'_minDoF',num2str(minNrDoF)];
    else
        folder = ['output/reduced_',example,'_Q',Q,'_alpha',num2str(alpha),'_minDoF',num2str(minNrDoF),'_theta',num2str(theta)];
    end
    % Make folder
    if ~exist(folder,'dir')
        mkdir(folder);
    end
    
    % Set number of nodes up to which mesh data should be saved
    triangBound = 10000;
    
    % Initialize data vectors
    nrDoF4lvl = [];
    eta4lvl = [];
    data4lvl = [];
    error4lvlTotal = [];
    error4lvlU = [];
    error4lvlV = [];
    
    if compare
        nrDoF4lvlDPG = [];
        eta4lvlDPG = [];
        data4lvlDPG = [];
        error4lvlTotalDPG = [];
        error4lvlRDPG = [];
        error4lvlWDPG = [];
        error4lvlTDPG = [];
        error4lvlUDPG = [];
        error4lvlVDPG = [];
        error4lvlQDPG = [];
    end
    
    %% Start AFEM loop SOLVE -> ESTIMATE -> MARK -> REFINE
    while(true)
        %%%%%%%%%%%%
        % SOLVE
        %%%%%%%%%%%%

        [v,u,nrDoF, dataReduced] = solveReducedFormulation(c4n,n4e,n4sDb,n4sNb,f,u4Db,g,alpha,Q);        
       
        nrDoF4lvl(end+1) = nrDoF;
        data4lvl = [data4lvl; dataReduced];
                    
        %%%%%%%%%%%%
        % ESTIMATE
        %%%%%%%%%%%%
        eta4e = estimateReduced(c4n,n4e,f,alpha,Q,v);
        % Collect estimator in vector
        eta4lvl(end+1) = sqrt(sum(eta4e));
        
        %%%%%%%%%%%%if (compare && alpha==1 && strcmp(Q,'identity'))
        % (Errors)
        %%%%%%%%%%%%
        % Compute errors if exact solution is known
        if exSolKnown == 1
            [error4eTotal,error4eU,error4eV] = error4eReduced(c4n,n4e,graduExact,u,v);
            % Collect errors in vectors
            error4lvlTotal(end+1) = sqrt(sum(error4eTotal));  
            error4lvlU(end+1) = sqrt(sum(error4eU));  
            error4lvlV(end+1) = sqrt(sum(error4eV));              
        end    
            
        % Output in terminal
        disp(['nodes/dofs: ',num2str(size(c4n,1)),'/',num2str(nrDoF),...
             '; estimator = ',num2str(eta4lvl(end))]); 
         
        % save data
        if exSolKnown == 1
            fileID = fopen([folder,'/convergence.dat'],'w');
            fprintf(fileID,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlU','error4lvlV','error4lvlTotal','eta4lvl', 'assemble', 'solve', 'postprocess', 'cond');
            dlmwrite([folder,'/convergence.dat'], [nrDoF4lvl',error4lvlU',error4lvlV',error4lvlTotal',eta4lvl',data4lvl],'delimiter', '\t', '-append')
            fclose(fileID);
            close all; 
        else  
            fileID = fopen([folder,'/convergence.dat'],'w');
            fprintf(fileID,'%s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','eta4lvl', 'assemble', 'solve', 'postprocess', 'cond');
            dlmwrite([folder,'/convergence.dat'], [nrDoF4lvl',eta4lvl',data4lvl],'delimiter', '\t', '-append')
            fclose(fileID);
        end 
         
        % If comparison is wanted, solve, estimate and compute errors for corresponding dPG methods 
        if (compare && alpha==1 && strcmp(Q,'identity'))
            [uDPG,tDPG,vDPG, nrDoFDPG, dataDPG] = solveDPGPrimal(c4n,n4e,n4sDb,n4sNb,f,u4Db,g);
            nrDoF4lvlDPG(end+1) = nrDoFDPG;
            data4lvlDPG = [data4lvlDPG; dataDPG];
            [eta4eDPG,error4eVDPG] = estimatePrimalDPG(c4n,n4e,f,vDPG); 
            eta4lvlDPG(end+1) = sqrt(sum(eta4eDPG)); 
            error4lvlVDPG(end+1) = sqrt(sum(error4eVDPG)); 
            fileIDDPG = fopen([folder,'/DPGTime.dat'],'w');
            fprintf(fileIDDPG,'%s\t %s\t %s\t %s\n','nrDoF4lvl','assemble', 'solve', 'cond');
            dlmwrite([folder,'/DPGTime.dat'], [nrDoF4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
            fclose(fileIDDPG);
            close all; 
            if exSolKnown == 1
                [error4eUDPG,error4eTDPG] = error4ePrimalDPG(c4n,n4e,f,uExact,graduExact,uDPG,tDPG);
                error4lvlUDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTDPG(end+1) = sqrt(sum(error4eTDPG)); 
                error4lvlTotalDPG(end+1) = sqrt(sum(error4eVDPG+error4eUDPG+error4eTDPG));   
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlU','error4lvlT','error4lvlV','error4lvlTotal','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',error4lvlUDPG',error4lvlTDPG',error4lvlVDPG',error4lvlTotalDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
                close all; 
            else
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
            end
        elseif (compare && alpha==0.5 && strcmp(Q,'identity'))
            [rDPG,wDPG,tDPG,uDPG,qDPG,vDPG, nrDoFDPG, dataDPG] = solveDPGUltraweak(c4n,n4e,n4sDb,n4sNb,f,u4Db,g);
            nrDoF4lvlDPG(end+1) = nrDoFDPG;
            data4lvlDPG = [data4lvlDPG; dataDPG];
            [eta4eDPG,error4eQDPG,error4eVDPG] = estimateUltraweakDPG(c4n,n4e,f,qDPG,vDPG); 
            eta4lvlDPG(end+1) = sqrt(sum(eta4eDPG)); 
            error4lvlQDPG(end+1) = sqrt(sum(error4eQDPG));
            error4lvlVDPG(end+1) = sqrt(sum(error4eVDPG)); 
            fileIDDPG = fopen([folder,'/DPGTime.dat'],'w');
            fprintf(fileIDDPG,'%s\t %s\t %s\t %s\n','nrDoF4lvl','assemble', 'solve', 'cond');
            dlmwrite([folder,'/DPGTime.dat'], [nrDoF4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
            fclose(fileIDDPG);
            close all; 
            if exSolKnown == 1
                [error4eRDPG, error4eWDPG, error4eTDPG,error4eUDPG] = error4eUltraweakDPG(c4n,n4e,f,uExact,graduExact,rDPG,wDPG,tDPG,uDPG);
                error4lvlRDPG(end+1) = sqrt(sum(error4eRDPG)); 
                error4lvlWDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTDPG(end+1) = sqrt(sum(error4eTDPG)); 
                error4lvlUDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTotalDPG(end+1) = sqrt(sum(error4eVDPG+error4eQDPG+error4eRDPG+error4eUDPG+ error4eUDPG+error4eTDPG));   
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlR','error4lvlW','error4lvlT','error4lvlU','error4lvlQ','error4lvlV','error4lvlTotal','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',error4lvlRDPG',error4lvlWDPG',error4lvlTDPG',error4lvlUDPG',error4lvlQDPG',error4lvlVDPG',error4lvlTotalDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
                close all; 
            else
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
            end
        end 

        % Quit loop if minimal desired number of dof is reached
        if nrDoF >= minNrDoF 
            break;
        end
            
        %%%%%%%%%%%%%%%%%%%
        % MARK and REFINE
        %%%%%%%%%%%%%%%%%%%
        
        % Collective Marking algorithm
        if theta == 1
            % Compute new triangulation via uniform red refinement
            [c4n,n4e,n4sDb,n4sNb] = refineUniformRed(c4n,n4e,n4sDb,n4sNb);     
        else
            % Mark elements via bulk criterion
            n4sMarked = markBulk(n4e,eta4e,theta);
            % Compute new triangulation via Bi3/Green/Blue refinement
            [c4n,n4e,n4sDb,n4sNb] = refineBi3GB(c4n,n4e,n4sDb,n4sNb,n4sMarked);
        end
    end

    % For raw meshes, save mesh data
    if( size(c4n,1) <= triangBound)
    	fileID1 = fopen([folder,'/c4n.dat'],'w');
    	fprintf(fileID1,'%s\t  %s\n','x','y');
    	dlmwrite([folder,'/c4n.dat'], c4n,'delimiter', '\t', '-append')
    	fclose(fileID1);
    
    	fileID2 = fopen([folder,'/n4e.dat'],'w');
    	fprintf(fileID2,'%s\t  %s\t %s\n','P1','P2','P3');
    	dlmwrite([folder,'/n4e.dat'], n4e-1,'delimiter', '\t', '-append')
    	fclose(fileID2);
    end
    
    % save convergence history data and plot it
    if exSolKnown == 1
    % for known exact solution, errors can be plotted, too        
        % plot convergence to figure
        figure(1);
        plotConvergenceColor(nrDoF4lvl,error4lvlTotal,'||x - x_h||');
        hold on;
        plotConvergenceColor(nrDoF4lvl,error4lvlU,'||u - u_C||');
        plotConvergenceColor(nrDoF4lvl,error4lvlV,'||v_{CR}||');
        plotConvergenceColor(nrDoF4lvl,eta4lvl,['eta']);
    else
    % if no exact solution is known, plot only estimator      
        % plot convergence to figure
        figure(1);
        plotConvergenceColor(nrDoF4lvl,eta4lvl,['eta']);
        hold on;
    end
    
    % save convergence history plot to file
    print( figure(1), '-dpng', [folder,'/ConvergencePlot.png']);

    
    
    % for raw meshes, plot deformation and save triangulation to file
    if ( size(c4n,1)<=triangBound )
         % plot u
         figure(2);
         plotP1(c4n, n4e, u(:), 'P1 Solution');

         % plot v
         figure(3);
         plotCR(c4n, n4e, v(:), 'Riesz Representation');

         % plot interpolation of exact solution
         if exSolKnown == 1
             uCompare = uExact(c4n);
             figure(4);
             plotP1(c4n, n4e, uCompare, 'Exact Solution');
         end
        
         % save triangulation to file
         figure(5);
         plotTriangulation(c4n,n4e);
         print( figure(5), '-dpng', [folder,'/Triangulation.png']);

         
         % save solution u to file
         fileID3 = fopen([folder,'/u.dat'],'w');
         fprintf(fileID3,'%s\t %s\t %s\n','x','y','u');
         dlmwrite([folder,'/u.dat'],[c4n(n4e',:),reshape(u(n4e'),3*size(n4e,1),1)],'delimiter','\t','-append');
         fclose(fileID3);
         
         % save solution v to file
         s4e = computeS4e ( n4e ) ;
         v4eSides = v(s4e )';
         v4eNodes = ( ones(3)-2*eye(3) ) * v4eSides ;
         % Save piecewise affine and globally discontinuous scalar function u
         fileID4 = fopen([folder,'/v.dat'],'w');
         dlmwrite ( [folder,'/v.dat'],[c4n(n4e(:,[3 1 2])',:),reshape(v4eNodes,3*size(n4e,1),1)],'delimiter','\t','newline','unix');
         fclose(fileID4);
    end
    
end



% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
