function afemWeightedLS(example,tag_M0,tag_F0,compare,theta,kappa,rho,minNrDoF)
% Function for running benchmarks for weighted LS method 
% Input:   example    problem to solve, for options see below
%          tag_M0     piecewise constant SPD matrix in weighted LS, for options see below
%          tag_F0     piecewise constant right hand side vector in weighted LS, for options see below
%          compare    compare to equivalent dPG (compare = 1) or not ( = 0)
%          theta      bulk parameter, uniform refinement if set to 1
%          kappa      parameter for separate marking
%          rho        parameter for separate marking
%          minNrDoF   number of degrees of freedom to be reached minimally
    
    % Add current path
    addpath(genpath(pwd));

    %% Set data if none is given
    if( nargin == 0)
        % Options for problem
%         example = 'LshapeDirac';        
        example = 'LshapeDiracReverse';        
%         example = 'LshapeExact';        
%         example = 'OneRHS';        
%         example = 'SlitExact';        
%         example = 'SquareExact';        
%         example = 'SquareExact2';        
%         example = 'SquareExact3';        
%         example = 'Waterfall';

        compare = 0;

%         tag_M0 = 'identity'; % M_0 = identity matrix on each triangle
        tag_M0 = 'identitySZero'; % M_0 = identity matrix + Pi_0((identity-mid)^2)
%         tag_M0 = 'twoidentitySZero'; % M_0 = 2* identity matrix + Pi_0((identity-mid)^2)

%         tag_F0 = 'zero'; % F_0 = zero vector
        tag_F0 = 'H0'; % F_0 = Pi_0(f*(identity-mid))

        % Set Parameters
        theta = 0.5;
        kappa = 0.005;
        rho = 0.1;
        minNrDoF = 2300; 
    end

    %% Collect data for example
    data = eval(example);
    c4n = data{1};
    n4e = data{2};
    n4sDb = data{3};
    n4sNb = data{4};
    f = data{5};
    u4Db = data{6};
    g = data{7};
    exSolKnown = data{8};
    % optional exact solution data
    if exSolKnown == 1
        uExact = data{9};
        graduExact = data{10}; 
    end 
    
    % Set type of right-hand side approximation for weighted LS as ||(1-Pi_0)f||
    tag_approx = '1-piZero';
    
    
    % Set name for folder in which to save data
    if theta == 1
        folder = ['output/weightedLS_',example,'_M',tag_M0,'_F',tag_F0,'_minDoF',num2str(minNrDoF)];
    else
        folder = ['output/weightedLS_',example,'_M',tag_M0,'_F',tag_F0,'_minDoF',num2str(minNrDoF),'_theta',num2str(theta),'_kappa',num2str(kappa),'_rho',num2str(rho)];
    end
    % Make folder
    if ~exist(folder,'dir')
        mkdir(folder);
    end
    
    % Set number of nodes up to which mesh data should be saved
    triangBound = 10000;
    
    
    % Save initial triangulation for later use in approx algorithm
    c4n_initial = c4n;
    n4e_initial = n4e;
    n4sDb_initial = n4sDb;
    n4sNb_initial = n4sNb;
    % Compute right-hand side approximation for initial triangulation for
    % later use in approx algorithm
    mu4e_initial = calculateApproximationRHS(c4n,n4e,f,tag_approx);
    err4e_initial = [mu4e_initial, mu4e_initial];
    
    % For first call of approx algorithm, the triangulation from the last
    % call is the initial triangulation
    c4n_before = c4n_initial;
    n4e_before = n4e_initial;
    n4sDb_before = n4sDb_initial;
    n4sNb_before = n4sNb_initial;
    err4e_before = err4e_initial;
    
    % Initialize data vectors
    nrDoF4lvl = [];
    eta4lvl = [];
    mu4lvl = [];
    sigma4lvl = [];
    data4lvl = [];
    refine4lvl = [];
    case4lvl = [];
    error4lvlTotal = [];
    error4lvlP = [];
    error4lvlU = [];
    
    if compare
        nrDoF4lvlDPG = [];
        eta4lvlDPG = [];
        data4lvlDPG = [];
        error4lvlTotalDPG = [];
        error4lvlRDPG = [];
        error4lvlWDPG = [];
        error4lvlTDPG = [];
        error4lvlUDPG = [];
        error4lvlVDPG = [];
        error4lvlQDPG = [];
    end
    
    %% Start AFEM loop SOLVE -> ESTIMATE -> MARK -> REFINE with Separate Marking
    while(true)
        %%%%%%%%%%%%
        % SOLVE
        %%%%%%%%%%%%
        [p,u,nrDoF,dataLS] = solveWeightedLSFormulation(c4n,n4e,n4sDb,n4sNb,f,u4Db,g,tag_M0,tag_F0);
        nrDoF4lvl(end+1) = nrDoF;
        data4lvl = [data4lvl; dataLS];
        
        %%%%%%%%%%%%
        % ESTIMATE
        %%%%%%%%%%%%
        eta4e = estimateWeightedLS(c4n,n4e,f,tag_M0,tag_F0,u,p);
        % Collect estimator in vector
        eta4lvl(end+1) = sqrt(sum(eta4e));
        
        % Compute right-hand side approximation
        mu4e = calculateApproximationRHS(c4n,n4e,f,tag_approx);
        % Collect approximation in vector
        mu4lvl(end+1) = sqrt(sum(mu4e));
        % Collect total estimator in vector
        sigma4lvl(end+1) = sqrt(sum(mu4e+eta4e));
        
        %%%%%%%%%%%%
        % (Errors)
        %%%%%%%%%%%%
        % Compute errors if exact solution is known
        if exSolKnown == 1
            [error4eTotal,error4eU,error4eP] = error4eWeightedLS(c4n,n4e,f,graduExact,u,p);
            % Collect errors in vectors
            error4lvlTotal(end+1)  = sqrt(sum(error4eTotal));  
            error4lvlU(end+1) = sqrt(sum(error4eU));  
            error4lvlP(end+1) = sqrt(sum(error4eP));              
        end    
            
        % Output in terminal
        if theta == 1
            disp(['nodes/dofs: ',num2str(size(c4n,1)),'/',num2str(nrDoF),...
             '; estimator = ',num2str(sigma4lvl(end))]); 
        else
            if mu4lvl(end)^2 <= kappa * eta4lvl(end)^2
                disp(['Case A, nodes/dofs: ',num2str(size(c4n,1)),'/',num2str(nrDoF),...
                 '; estimator = ',num2str(sigma4lvl(end))]);
                case4lvl = [case4lvl,1];
            else 
                disp(['Case B, nodes/dofs: ',num2str(size(c4n,1)),'/',num2str(nrDoF),...
                 '; estimator = ',num2str(sigma4lvl(end))]);  
                case4lvl = [case4lvl,2];
            end
        end
        
        % save data
        if exSolKnown == 1
            fileID = fopen([folder,'/convergence.dat'],'w');
            fprintf(fileID,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlU','error4lvlP','error4lvlTotal','sigma4lvl','eta4lvl','mu4lvl', 'assemble', 'solve', 'postprocess', 'cond', 'case4lvl');
            dlmwrite([folder,'/convergence.dat'], [nrDoF4lvl',error4lvlU',error4lvlP',error4lvlTotal',sigma4lvl',eta4lvl',mu4lvl',data4lvl,case4lvl'],'delimiter', '\t', '-append')
            fclose(fileID);
            close all; 
        else  
            fileID = fopen([folder,'/convergence.dat'],'w');
            fprintf(fileID,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','sigma4lvl','eta4lvl','mu4lvl', 'assemble', 'solve', 'postprocess', 'cond','case4lvl');
            dlmwrite([folder,'/convergence.dat'], [nrDoF4lvl',sigma4lvl',eta4lvl',mu4lvl',data4lvl,case4lvl'],'delimiter', '\t', '-append')
            fclose(fileID);
        end 
    
        % If comparison is wanted, solve, estimate and compute errors for corresponding dPG methods 
        if (compare && strcmp(tag_M0,'identitySZero') && strcmp(tag_F0,'H1'))
            [uDPG,tDPG,vDPG, nrDoFDPG, dataDPG] = solveDPGPrimal(c4n,n4e,n4sDb,n4sNb,f,u4Db,g);
            nrDoF4lvlDPG(end+1) = nrDoFDPG;
            data4lvlDPG = [data4lvlDPG; dataDPG];
            [eta4eDPG,error4eVDPG] = estimatePrimalDPG(c4n,n4e,f,vDPG); 
            eta4lvlDPG(end+1) = sqrt(sum(eta4eDPG)); 
            error4lvlVDPG(end+1) = sqrt(sum(error4eVDPG)); 
            fileIDDPG = fopen([folder,'/DPGTime.dat'],'w');
            fprintf(fileIDDPG,'%s\t %s\t %s\t %s\n','nrDoF4lvl','assemble', 'solve', 'cond');
            dlmwrite([folder,'/DPGTime.dat'], [nrDoF4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
            fclose(fileIDDPG);
            if exSolKnown == 1
                [error4eUDPG,error4eTDPG] = error4ePrimalDPG(c4n,n4e,f,uExact,graduExact,uDPG,tDPG);
                error4lvlUDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTDPG(end+1) = sqrt(sum(error4eTDPG)); 
                error4lvlTotalDPG(end+1) = sqrt(sum(error4eVDPG+error4eUDPG+error4eTDPG));   
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlU','error4lvlT','error4lvlV','error4lvlTotal','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',error4lvlUDPG',error4lvlTDPG',error4lvlVDPG',error4lvlTotalDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
                close all; 
            else
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
            end
        elseif (compare && strcmp(tag_M0,'twoidentitySZero') && strcmp(tag_F0,'H1'))
            [rDPG,wDPG,tDPG,uDPG,qDPG,vDPG, nrDoFDPG, dataDPG] = solveDPGUltraweak(c4n,n4e,n4sDb,n4sNb,f,u4Db,g);
            nrDoF4lvlDPG(end+1) = nrDoFDPG;
            data4lvlDPG = [data4lvlDPG; dataDPG];
            [eta4eDPG,error4eQDPG,error4eVDPG] = estimateUltraweakDPG(c4n,n4e,f,qDPG,vDPG); 
            eta4lvlDPG(end+1) = sqrt(sum(eta4eDPG)); 
            error4lvlQDPG(end+1) = sqrt(sum(error4eQDPG));
            error4lvlVDPG(end+1) = sqrt(sum(error4eVDPG)); 
            fileIDDPG = fopen([folder,'/DPGTime.dat'],'w');
            fprintf(fileIDDPG,'%s\t %s\t %s\t %s\n','nrDoF4lvl','assemble', 'solve', 'cond');
            dlmwrite([folder,'/DPGTime.dat'], [nrDoF4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
            fclose(fileIDDPG);
            if exSolKnown == 1
                [error4eRDPG, error4eWDPG, error4eTDPG,error4eUDPG] = error4eUltraweakDPG(c4n,n4e,f,uExact,graduExact,rDPG,wDPG,tDPG,uDPG);
                error4lvlRDPG(end+1) = sqrt(sum(error4eRDPG)); 
                error4lvlWDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTDPG(end+1) = sqrt(sum(error4eTDPG)); 
                error4lvlUDPG(end+1) = sqrt(sum(error4eUDPG)); 
                error4lvlTotalDPG(end+1) = sqrt(sum(error4eVDPG+error4eQDPG+error4eRDPG+error4eUDPG+ error4eUDPG+error4eTDPG));   
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','error4lvlR','error4lvlW','error4lvlT','error4lvlU','error4lvlQ','error4lvlV','error4lvlTotal','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',error4lvlRDPG',error4lvlWDPG',error4lvlTDPG',error4lvlUDPG',error4lvlQDPG',error4lvlVDPG',error4lvlTotalDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
                close all; 
            else
                fileIDDPG = fopen([folder,'/convergenceDPG.dat'],'w');
                fprintf(fileIDDPG,'%s\t %s\t %s\t %s\t %s\n','nrDoF4lvl','eta4lvl', 'assemble', 'solve', 'cond');
                dlmwrite([folder,'/convergenceDPG.dat'], [nrDoF4lvlDPG',eta4lvlDPG',data4lvlDPG],'delimiter', '\t', '-append')
                fclose(fileIDDPG);
            end
        end 
        
        
        % Quit loop if minimal desired number of dof is reached
        if nrDoF >= minNrDoF 
            break;
        end
            
        
        %%%%%%%%%%%%%%%%%%%
        % MARK and REFINE
        %%%%%%%%%%%%%%%%%%%
        refineStart = tic; % Measure the time spent for refinement (case A or B)        
        if theta == 1
            % Compute new triangulation via uniform red refinement
            [c4n,n4e,n4sDb,n4sNb] = refineUniformRed(c4n,n4e,n4sDb,n4sNb);     
        else
            % Separate Marking algorithm
            % Distinguish cases A and B
            %% Case A (Bulk Criterion)
            if mu4lvl(end)^2 <= kappa * eta4lvl(end)^2
                % Mark elements via bulk criterion
                n4sMarked = markBulk(n4e,eta4e,theta);
                % Compute new triangulation via Bi3/Green/Blue refinement
                [c4n,n4e,n4sDb,n4sNb] = refineBi3GB(c4n,n4e,n4sDb,n4sNb,n4sMarked);
            %% Case B (Approximation of right hand side)
            else
                % Compute triangulation such that new mu gets reduced to less than rho*mu4nrDoF(nrDoF)
                [c4n_approx, n4e_approx, n4sDb_approx, n4sNb_approx, err4e_approx, c4n_before, n4e_before, ...
                    n4sDb_before, n4sNb_before, err4e_before] = ...
                    approx(f, tag_approx, c4n_before, n4e_before, n4sDb_before, n4sNb_before, err4e_before,...
                     c4n_initial, n4e_initial, n4sDb_initial, n4sNb_initial, err4e_initial, rho^2*mu4lvl(end)^2);
                % Compute new triangulation via overlay of triangulation and
                % triangulation computed in previous step
                [c4n, n4e, n4sDb, n4sNb] = overlay(c4n_initial, n4e_initial, n4sDb_initial, ...
                            n4sNb_initial, c4n, c4n_approx);
            end
        end
        refine4lvl(end+1) = toc(refineStart);

        fileIDDPG = fopen([folder,'/RefineTime.dat'],'w');
        fprintf(fileIDDPG,'%s\t %s\n','nrDoF4lvl','refine4lvl');
        dlmwrite([folder,'/RefineTime.dat'], [nrDoF4lvl',refine4lvl'],'delimiter', '\t', '-append')
        fclose(fileIDDPG);
    end

    % For raw meshes, save mesh data
    if( size(c4n,1) <= triangBound)
    	fileID1 = fopen([folder,'/c4n.dat'],'w');
    	fprintf(fileID1,'%s\t  %s\n','x','y');
    	dlmwrite([folder,'/c4n.dat'], c4n,'delimiter', '\t', '-append')
    	fclose(fileID1);
    
    	fileID2 = fopen([folder,'/n4e.dat'],'w');
    	fprintf(fileID2,'%s\t  %s\t %s\n','P1','P2','P3');
    	dlmwrite([folder,'/n4e.dat'], n4e-1,'delimiter', '\t', '-append')
    	fclose(fileID2);
    end
     
    % save convergence history data and plot it
    if exSolKnown == 1
    % for known exact solution, errors can be plotted, too        
        % plot convergence to figure
        figure(1);
        plotConvergenceColor(nrDoF4lvl,error4lvlTotal,'||x - x_h||');
        hold on;
        plotConvergenceColor(nrDoF4lvl,error4lvlU,'||u - u_C||');
        plotConvergenceColor(nrDoF4lvl,error4lvlP,'||\nabla u - p_{RT}||');
        plotConvergenceColor(nrDoF4lvl,sigma4lvl,['sigma']);
        plotConvergenceColor(nrDoF4lvl,eta4lvl,['eta']);
        plotConvergenceColor(nrDoF4lvl,mu4lvl,['mu']);
    else
    % if no exact solution is known, plot only estimator
        % plot convergence to figure
        figure(1);
        plotConvergenceColor(nrDoF4lvl,sigma4lvl,['sigma']);
        hold on;
        plotConvergenceColor(nrDoF4lvl,eta4lvl,['eta']);
        plotConvergenceColor(nrDoF4lvl,mu4lvl,['mu']);
    end
    
    % save convergence history plot to file
    print( figure(1), '-dpng', [folder,'/ConvergencePlot.png']);

    
    
    % for raw meshes, plot deformation and save triangulation to file
    if ( size(c4n,1)<=triangBound )
         figure(2);
         plotP1(c4n, n4e, u(:), 'P1 Solution');

         if exSolKnown == 1
             uCompare = uExact(c4n);
             figure(3);
             plotP1(c4n, n4e, uCompare, 'Exact Solution');
         end
         
         % save triangulation to file
         figure(4);
         plotTriangulation(c4n,n4e);
         print( figure(4), '-dpng', [folder,'/Triangulation.png']);

         % save solution u to file
         fileID3 = fopen([folder,'/u.dat'],'w');
         fprintf(fileID3,'%s\t %s\t %s\n','x','y','u');
         dlmwrite([folder,'/u.dat'],[c4n(n4e',:),reshape(u(n4e'),3*size(n4e,1),1)],'delimiter','\t','-append');
         fclose(fileID3);
         
         
         pCompare = zeros(size(n4e,1),2);
         n4s = computeN4s(n4e);
         area4e = computeArea4e(c4n,n4e);
         mid4e = computeMid4e(c4n, n4e);
         length4s = computeLength4s(c4n,n4s);
         normal4e = computeNormal4e(c4n,n4e);
         normal4s = computeNormal4s(c4n,n4s);
         s4e = computeS4e(n4e);
         for curElem = 1 : size(n4e,1)
            % Compute the coordinates of the nodes of the current element
            coords = c4n(n4e(curElem,:),:);
            % Compute the sides of the current element in global enumeration of
            % sides as in n4s
            sides = s4e(curElem,:);
            signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
            pCompare(curElem,:) =sum(repmat(p(sides,:).*signedLength',1,2).*(repmat(mid4e(curElem,:),3,1)-coords([3 1 2],:)))/(2*area4e(curElem));
         end
         % save solution t to file (quiver plot)
         fileID6 = fopen([folder,'/p.dat'],'w');
         fprintf(fileID6,'%s\t %s\t %s\t %s\t %s\n','x','y','r1','r2','z');
         dlmwrite([folder,'/p.dat'],[mid4e,pCompare,sqrt(sum(pCompare.^2,2))],'delimiter','\t','-append');
         fclose(fileID6); 
    end
   
    figure(3);
    plotP1(c4n, n4e, u(:), 'P1 Solution');
end


% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

