function eta4e = estimateReduced(c4n,n4e,f,alpha,Q,vApprox)
% Function to compute estimator for reduced mixed method
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         f	       right hand side function
%         alpha    real parameter in reduced formulation
%         Q        local projection in reduced formulation
%         vApprox  Crouzeix Raviart approximation v_{CR} of residual
% Output: eta4e    error estimator for each element

	%% Compute mesh quantities
    s4e      = computeS4e(n4e);
    n4s      = computeN4s(n4e);
    length4s = computeLength4s(c4n,n4s);
    mid4e = computeMid4e(c4n,n4e);
    area4e = computeArea4e(c4n,n4e);
    normal4e = computeNormal4e(c4n,n4e);
    normal4s = computeNormal4s(c4n,n4s);
    
    % Compute L^2 products with projections Q
    [Qproductlocal,RHSproductlocal] = calculateL2Product(c4n,n4e,f,Q);
    
    degree = 12;
    % Compute integral of f^2 on each element                           
    intFSq4e = integrate(c4n,n4e,@(n4p,Gpts4p,Gpts4ref)sum(f(Gpts4p).^2,2),degree,area4e);
    
    % Prepare vector for gradients of vApprox and oscillations
    gradV = zeros(size(n4e,1),2);
    osc4e = zeros(size(n4e,1), 1);
    
    %% Compute gradients of vApprox and oscillations
    parfor curElem = 1:size(n4e,1);
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the barycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Gradients of CR basis are scaled gradients of barycentric
        % coordinates
        grads = -2*grads([3 1 2],:);
        gradV(curElem,:) = vApprox(s4e(curElem,:))' * grads;
        
        % Compute oscillations || f - alpha Q vApprox ||^2 via the split
        % || f - alpha Q vApprox ||^2 = ||f||^2 - 2alpha (f,Q vApprox) +...
        %  alpha^2 || Q vApprox ||^2
        osc4e(curElem) = intFSq4e(curElem) - ...
            2*alpha*vApprox(s4e(curElem,:))'*RHSproductlocal(:,curElem) + ... 
            alpha^2 * vApprox(s4e(curElem,:))'* Qproductlocal(:,:,curElem)*vApprox(s4e(curElem,:));
    end
    
    %% Compute the L2-norm of the jumps.
    jump4s = zeros(size(n4s,1),2);
    for curElem = 1 : size(n4e,1)
        sides = s4e(curElem,:);
        signs = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)');
        jump4s(s4e(curElem,:),:) = jump4s(s4e(curElem,:),:) ...
                              + signs'*gradV(curElem,:);
    end
    eta4s = sum(jump4s.^2,2).*length4s;

    %% Compute total error estimator 
    eta4e = area4e.*osc4e + sqrt(area4e).*sum(eta4s(s4e),2);
end



% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.