function eta4e = estimateWeightedLS(c4n,n4e,f,tag_M0,tag_F0,uApprox,pApprox)
% Function to compute estimator for weighted LS formulation
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         f	       right hand side function
%         tag_M0   tag for choice of M_0 in weighted LS
%         tag_F0   tag for choice of F_0 in weighted LS
%         uApprox  Courant approximation u_C of solution
%         pApprox  Raviart-Thomas approximation p_{RT} of \nabla u
% Output: eta4e    error estimator for each element

	%% Compute mesh quantities
    s4e      = computeS4e(n4e);
    n4s      = computeN4s(n4e);
    length4s = computeLength4s(c4n,n4s);
    mid4e = computeMid4e(c4n,n4e);
    area4e = computeArea4e(c4n,n4e);
    normal4e = computeNormal4e(c4n,n4e);
    normal4s = computeNormal4s(c4n,n4s);

    % Prepare vectors for jump terms and oscillations
    jumpTerm = zeros(size(n4e,1),2);
    oscP4e = zeros(size(n4e,1), 1);
    
    % Compute weight function M_0 (piecewise constant 2x2 SPD matrix)
    M0 = calculateM0(c4n,n4e,tag_M0);
    % Compute right hand side oscillation (piecewise constant 2-vector)
    F0 = calculateF0(c4n,n4e,tag_F0,f);

    %% Compute gradient of u and projection of p.
    parfor curElem = 1:size(n4e,1);
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the braycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Compute the sides of the current element in global enumeration of
        % sides as in n4s
        sides = s4e(curElem,:);
        % Compute the vector containing the lengths of the three sides of
        % the current element, equipped with a sign correspnding to the
        % orientation of the current side: 1 if element-normal equals to
        % edge-normal, -1 if the normals are directed in opposite
        % directions
        signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
        
        % Compute a matrix containing the vectors of the three edges
        edges = coords - coords([2,3,1],:);
        % Compute the local norm of div p_{RT}(identity-mid)/2
        oscP4e(curElem) = area4e(curElem)*sum(signedLength)^2* trace(edges'*edges)/144;
        
        % Compute matrix representing the piecewise constant parts of the
        % three Raviart-Thomas basis functions of the edges
        Q = 1/(2*area4e(curElem))*repmat(signedLength',1,2).* (repmat(mid4e(curElem,:),3,1)-coords([3,1,2],:));

        % Compute the jump term consisting of the sum of
        % gradient of uApprox +
        % piecewise constant part of pApprox +
        % right hand side oscillation F0
        % all weighted with M_0
        jumpTerm(curElem,:) = (pApprox(s4e(curElem,:))'*Q - uApprox(n4e(curElem,:))'* grads + F0(curElem,:))*inv(M0(:,:,curElem));
    end
        
    
    %% Compute the L2-norm of the jumps.
    jump4s = zeros(size(n4s,1),2);
    for curElem = 1 : size(n4e,1)
        sides = s4e(curElem,:);
        signs = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)');
        jump4s(s4e(curElem,:),:) = jump4s(s4e(curElem,:),:) ...
                              + signs'*jumpTerm(curElem,:);
    end
    eta4s = sum(jump4s.^2,2).*length4s;
    

    %% Output 
    eta4e = oscP4e + sqrt(area4e).*sum(eta4s(s4e),2);
end



% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.