function mu4e = calculateApproximationRHS(c4n,n4e,f,approx)
% Function to compute approximation errors mu^2(T) on any element
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         f	       right hand side function
%         approx   tag for type of approximation error
% Output: mu4e     vector with values mu^2

    %% Distinguish cases for different possibilities of approximation error
    % Case for approx = oscillation, i.e., mu^2(T) = |T| ||(1-Pi_0)f||_{L^2(T)}^2
    if strcmp(approx,'oscillation')
        mu4e = calculateApproximationRHSOscillation(c4n,n4e,f);
    % Case for approx = 1-piZero, i.e. mu^2(T) = ||(1-Pi_0)f||_{L^2(T)}^2
    elseif strcmp(approx,'1-piZero')
        mu4e = calculateApproximationRHSPiZero(c4n,n4e,f);
    else 
        disp('approximation error not defined.'); 
    end
end

%% Case for approx = oscillation
function mu4e = calculateApproximationRHSOscillation(c4n,n4e,f);
    area4e = computeArea4e(c4n,n4e); 
    % Compute oscillation as ||(1-Pi_0)f|| weighted with the mesh size
    mu4e = area4e.*calculateApproximationRHSPiZero(c4n,n4e,f);
end

%% Case for approx = ||(1-Pi_0)f||_T^2
function mu4e = calculateApproximationRHSPiZero(c4n,n4e,f);
    %% Compute mesh quantities
    area4e = computeArea4e(c4n,n4e);
    
    % degree of numerical integration
    degreeF = 12; 

    % Compute integral of f^2 on each element
    intFSq4e = integrate(c4n, n4e, @(n4p,Gpts4p,Gpts4ref)((f(Gpts4p)).^2), ...
                   degreeF, area4e);
    
    % Compute integral mean of f on each element
    ProjF = integrate(c4n, n4e, @(n4p,Gpts4p,Gpts4ref)f(Gpts4p), ...
                   degreeF, area4e)./area4e;  
               
    % Compute ||(1-Pi_0)f||^2 = ||f||^2 - ||Pi_0f||^2           
    mu4e = intFSq4e - area4e.* ProjF.^2;
end


% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
