function [localMassMatrices,localRHS] = calculateL2Product(c4n,n4e,f,Q)
% Function to compute local mass matrices for Crouzeix-Raviart basis and 
% scalar product (.,Q.) and local right-hand side matrices for functional (f,Q.)
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         f	       right hand side function
%         Q        tag for local projection in reduced formulation
% Output: localMassMatrices   3x3 matrix for each element
%         localRHS            3-vector for each element

    %% Distinguish cases for different possibilities of projection Q
    % Case for Q = identity mapping
    if strcmp(Q,'identity')
        [localMassMatrices,localRHS] = calculateL2ProductId(c4n,n4e,f);
    % Case for Q = L^2 projection onto piecewise constants
    elseif strcmp(Q,'piZero')
        [localMassMatrices,localRHS] = calculateL2ProductPi(c4n,n4e,f);
    else 
        disp('Q not defined.'); 
    end
end

%% Case for Q = identity mapping
function [localMassMatrices,localRHS] = calculateL2ProductId(c4n,n4e,f)

    % Compute mesh quantities
    nrElems = size(n4e,1);
    area4e = computeArea4e(c4n,n4e);
    
    % Degree of numerical integration
    degree = 20;
    
    % Prepare output vectors
    localMassMatrices = zeros(3,3,nrElems);
    localRHS = zeros(3,nrElems);
    
    % Initialize function handles for barycentric coordinates
    [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e);
    % Integrals of the products of f with the
    % local barycentric coordinates on each element
    intFLambdaOne4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda1(Gpts4p)),degree,area4e);
    intFLambdaTwo4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda2(Gpts4p)),degree,area4e);
    intFLambdaThree4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda3(Gpts4p)),degree,area4e);
    intFLambda4e = [intFLambdaOne4e, intFLambdaTwo4e, intFLambdaThree4e]; 
    
%     parfor curElem = 1:nrElems
    parfor curElem = 1:nrElems
        % mass matrix of (.,.)_{L^2} for CR is 1/3* identity matrix
        localMassMatrices(:,:,curElem) = area4e(curElem)/3*eye(3);
        % use midpoint quadrature to evaluate (f,w_{CR})
        localRHS(:,curElem) = intFLambda4e(curElem,:)';
    end
end

%% Case for Q = L^2 projection onto piecewise constants
function [localMassMatrices,localRHS] = calculateL2ProductPi(c4n,n4e,f)
    % Compute mesh quantities
    nrElems = size(n4e,1);
    area4e = computeArea4e(c4n,n4e);
    
    % Degree of numerical integration
    degree = 20;
    
    % Prepare output vectors
    localMassMatrices = zeros(3,3,nrElems);
    localRHS = zeros(3,nrElems);
    
    % Compute integral of f on each element
    intF4e = integrate(c4n,n4e,@(n4p,Gpts4p,Gpts4ref)(f(Gpts4p)),degree,area4e);
    
    parfor curElem = 1:nrElems
        % mass matrix of (Pi_0 .,.)_{L^2} for CR is 1/9*ones(3,3)
        localMassMatrices(:,:,curElem) = area4e(curElem)/9*ones(3,3);
        % use midpoint quadrature to evaluate (Pi_0 f,w_{CR})
        localRHS(:,curElem) = 1/3 * intF4e(curElem,:)*ones(3,1);
    end
end

function [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e)
        %% Compute vertices
        P1 = c4n(n4e(:,1),:);
        P2 = c4n(n4e(:,2),:);
        P3 = c4n(n4e(:,3),:);
        
        
        %% Define barycentric coordinates
        % by means of affine transformation Phi x + P3
        % where Phi = [P1-P3,P2-P3] '2x2-matrix' in row notation
        % where [a c; b d] --> [a b c d]
        Phi = [P1-P3,P2-P3];
        % Compute Psi = inverse of Phi
        Psi = repmat(1./(Phi(:,1).*Phi(:,4) - Phi(:,2).*Phi(:,3)),1,4)...
                 .*[Phi(:,4),-Phi(:,2),-Phi(:,3),Phi(:,1)];
        % Barycentric coordinates
        Lambda1 = @(z)(Psi(:,1).*(z(:,1)-P3(:,1)) + Psi(:,3).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda2 = @(z)(Psi(:,2).*(z(:,1)-P3(:,1)) + Psi(:,4).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda3 = @(z)(1 - Lambda1(z) - Lambda2(z));     
end


% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.