function [error4eU,error4eT] = error4ePrimalDPG(c4n,n4e,f,graduExact,uApprox,tApprox)
% Function to compute errors for primal dPG method for problems where exact solution is given
% Input:  c4n           coordinates for the nodes of the mesh
%         n4e           nodes for the elements of the mesh
%         f	            right hand side function
%         graduExact    gradient of exact solution
%         uApprox       Courant approximation u_C of solution
%         tApprox       P_0 approximation on edges of normal trace
%                           of grad u
% Output: error4eU      error component of u_C in H^1 norm
%         error4eT      error component of RT-extension of t in H(div) norm

    %% Compute mesh quantities
    nrElems = size(n4e,1);
    s4e = computeS4e(n4e);
    area4e = computeArea4e(c4n,n4e);
    n4s = computeN4s(n4e);
    length4s = computeLength4s(c4n,n4s);
    mid4e = computeMid4e(c4n, n4e);
    normal4e = computeNormal4e(c4n,n4e);
    normal4s = computeNormal4s(c4n,n4s);

    % Degree of numerical integration
    degree = 12;

    %% Preparations for computation of error in p
    % Initialize function handles for barycentric coordinates
    [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e);
    % Compute integral of f on each element
    intF4e = integrate(c4n,n4e,@(n4p,Gpts4p,Gpts4ref)(f(Gpts4p)),degree,area4e);

    % Compute integral of f^2 on each element                           
    intFSq4e = integrate(c4n,n4e,@(n4p,Gpts4p,Gpts4ref)sum(f(Gpts4p).^2,2),degree,area4e);                
    % Integrals of the products of exact gradient with the
    % local barycentric coordinates on each element
    intGradLambdaOne4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(graduExact(Gpts4p).*...
                        [Lambda1(Gpts4p) Lambda1(Gpts4p)]),degree,area4e);
    intGradLambdaTwo4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(graduExact(Gpts4p).*...
                        [Lambda2(Gpts4p) Lambda2(Gpts4p)]),degree,area4e);
    intGradLambdaThree4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(graduExact(Gpts4p).*...
                        [Lambda3(Gpts4p) Lambda3(Gpts4p)]),degree,area4e);
    % Integral of squared norm of exact gradient
    intGradSq4e = integrate(c4n,n4e,@(n4p,Gpts4p,Gpts4ref)...
                        sum(graduExact(Gpts4p).*graduExact(Gpts4p),2),...
                        degree,area4e);
                    
    % Preparation of vectors collecting local quantities
    gradU = zeros(nrElems,2);
    error4eTL2 = zeros(nrElems,1);
    error4eTDiv = zeros(nrElems,1); 
    
    % Compute gradient of uApprox and 
    parfor curElem = 1 : nrElems
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the barycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Compute gradient of uApprox
        gradU(curElem,:) = uApprox(n4e(curElem,:),1)' * grads;
        % Compute the sides of the current element in global enumeration of
        % sides as in n4s
        sides = s4e(curElem,:);                      
        % Compute the vector containing the lengths of the three sides of
        % the current element, equipped with a sign correspnding to the
        % orientation of the current side: 1 if element-normal equals to
        % edge-normal, -1 if the normals are directed in opposite
        % directions
        signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
        % Compute matrix representing the piecewise constant parts of the
        % three Raviart-Thomas basis functions of the edges
        Q = 1/(2*area4e(curElem))*repmat(signedLength',1,2).* (repmat(mid4e(curElem,:),3,1)-coords([3,1,2],:));
        % Compute a matrix containing the vectors of the three edges
        edges = coords - coords([2,3,1],:);
        % Compute L^2 mass matrix of divergence for local RT basis
        Adiv = (1/area4e(curElem))*(signedLength'*signedLength);
        % Compute L^2 mass matrix for local RT basis 
        AL2 = area4e(curElem)*(Q*Q') + trace(edges'*edges)/144 * Adiv;
        
        % Coefficients of pApprox corresponding to the current element
        tApproxLoc = reshape(tApprox(sides,:),3,1);
        % Compute || f + div pApprox ||^2 = || f ||^2 + || div pApprox ||^2
        % + 2 (f, div pApprox)
        error4eTDiv(curElem) = intFSq4e(curElem) + tApproxLoc'*Adiv*tApproxLoc + 2*tApproxLoc'*intF4e(curElem,:)*signedLength'./area4e(curElem,:);
        
        % Compute vector containing L^2 norm of gradient of u multiplied 
        % with local RT basis
        intGradLambda = [intGradLambdaOne4e(curElem,:);...
                            intGradLambdaTwo4e(curElem,:);...
                            intGradLambdaThree4e(curElem,:)];
        gradURT = ones(1,3)*intGradLambda*Q' + (coords(:)'*intGradLambda(:) - ones(1,3)*intGradLambda*mid4e(curElem,:)')*signedLength/2;
        
        % Compute || nabla u - pApprox ||^2 = || nabla u ||^2 + || pApprox ||^2
        % - 2 (nabla u, pApprox)
        error4eTL2(curElem) = intGradSq4e(curElem) + tApproxLoc'*AL2*tApproxLoc - 2*gradURT*tApproxLoc;  
    end

    
    % Compute L2 error of gradient of u
    error4eU = integrate(c4n,n4e,@(n4p, Gpts4p, Gpts4ref) (...
        sum((graduExact(Gpts4p) - gradU ).^2,2)),degree);
    
    % Add L2 and divergence error of p
    error4eT = error4eTL2 + error4eTDiv;
end

function [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e)
        %% Compute vertices
        P1 = c4n(n4e(:,1),:);
        P2 = c4n(n4e(:,2),:);
        P3 = c4n(n4e(:,3),:);
        
        
        %% Define barycentric coordinates
        % by means of affine transformation Phi x + P3
        % where Phi = [P1-P3,P2-P3] '2x2-matrix' in row notation
        % where [a c; b d] --> [a b c d]
        Phi = [P1-P3,P2-P3];
        % Compute Psi = inverse of Phi
        Psi = repmat(1./(Phi(:,1).*Phi(:,4) - Phi(:,2).*Phi(:,3)),1,4)...
                 .*[Phi(:,4),-Phi(:,2),-Phi(:,3),Phi(:,1)];
        % Barycentric coordinates
        Lambda1 = @(z)(Psi(:,1).*(z(:,1)-P3(:,1)) + Psi(:,3).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda2 = @(z)(Psi(:,2).*(z(:,1)-P3(:,1)) + Psi(:,4).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda3 = @(z)(1 - Lambda1(z) - Lambda2(z));     
end






% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.