function [error4eTotal,error4eU,error4eV] = error4eReduced(c4n,n4e,graduExact,uApprox,vApprox)
% Function to compute errors for reduced mixed methods for problems where exact solution is given
% Input:  c4n           coordinates for the nodes of the mesh
%         n4e           nodes for the elements of the mesh
%         graduExact    gradient of exact solution
%         uApprox       Courant approximation u_C of solution
%         vApprox       Crouzeix Raviart approximation v_{CR} of residual
% Output: error4eTotal  total error
%         error4eU      error component of u_C in energy norm
%         error4eV      error component of v_{CR} in piecewise H^1 norm

    %% Compute mesh quantities
    nrElems = size(n4e,1);
    s4e = computeS4e(n4e);
    area4e = computeArea4e(c4n,n4e);

    % Prepare output field for error of vApprox
    error4eV = zeros(nrElems,1);
    % Vector for gradients of uApprox
    gradU = zeros(nrElems,2);

    degree = 12;

    
    %% Compute H^1 error between vApprox and v=0
%     parfor curElem = 1:nrElems
    parfor curElem = 1:nrElems
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the barycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Compute gradient of uApprox
        gradU(curElem,:) = uApprox(n4e(curElem,:),1)' * grads;
        % Gradients of CR basis are scaled gradients of barycentric
        % coordinates
        grads = -2*grads([3 1 2],:);
        % Compute H^1 norm of vApprox
        error4eV(curElem) = area4e(curElem) * vApprox(s4e(curElem,:))'*(grads * grads' + eye(3)/3)*vApprox(s4e(curElem,:));  
    end
    
    
    %% Compute error between u and u_C
    % Compute L2 error between gradients of u and u_C   
    error4eU = integrate(c4n,n4e,@(n4p, Gpts4p, Gpts4ref) (...
        sum((graduExact(Gpts4p) - gradU ).^2,2)),degree);
    
    %% Compute total error
    error4eTotal = error4eU + error4eV;

end

% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.