function [r,w,t,u,q, v,nrDoF, data] = solveDPGUltraweak(c4n,n4e,n4sDb,n4sNb,f,u4Db,g)
% Function to compute discrete solution to ultraweak dPG formulation for the
% Poisson model problem.
% The Poisson model problem seeks u with
%             - Delta u = f         on Omega
%                     u = u4Db      on Dirichlet boundary of Omega
%          nabla u * nu = g         on Neumann boundary of Omega
% given righthand side f, Dirichlet boundary condition u4Db and 
% Neumann boundary condition g on the domain Omega given by c4n and n4e.
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         n4sDb	   nodes of the sides in the Dirichlet boundary
%         n4sNb	   nodes of the sides in the Neumann boundary
%         f	       right hand side function
%         u4Db     Dirichlet boundary condition function
%         g        Neumann boundary condition function
% Output: r        vector valued piecewise P_0 approximation of nabla u 
%         w        piecewise P_0 approximation of u
%         t        P_0 approximation on edges of normal trace
%                      of grad u
%         u        globally continuous  P_1 approximation
%                      on edges of trace of u
%         q        piecewise RT_0 approximation of residual
%         v        piecewise P_1 approximation of residual
%         nrDof    number of degrees of freedom (size of system matrix)
%         data     data array containing time for solve, assembly and
%                      condition number of system matrix

    
    %% Preparation of variables, scaling
    % Numbers of elements, nodes and sides
    nrElems = size(n4e,1);
    nrNodes = size(c4n,1);
    n4s = computeN4s(n4e);
    nrSides = size(n4s,1);

    % Compute areas, mid points, lengths of sides, normals
    area4e = computeArea4e(c4n,n4e);
    mid4s = computeMid4s(c4n, n4s);
    mid4e = computeMid4e(c4n, n4e);
    length4s = computeLength4s(c4n,n4s);
    normal4e = computeNormal4e(c4n,n4e);
    normal4s = computeNormal4s(c4n,n4s);
    s4e = computeS4e(n4e);
    s4n = computeS4n(n4e);
    
    % Compute degrees of freedom for nodes
    nodes4Db = unique(n4sDb)';
    freeNodes = setdiff(1:nrNodes,nodes4Db);
    nrFreeNodes = size(freeNodes,2);
    
    % Compute degrees of freedom for sides
    sides4Nb = unique(s4n(sub2ind(size(s4n),n4sNb(:,1),n4sNb(:,2))))';
    freeSides = setdiff(1:nrSides, sides4Nb);
    nrFreeSides = size(freeSides,2);

    % Put together all degrees of freedom
    DoF = [1:(9*nrElems), 9*nrElems + freeSides, 9*nrElems + nrSides + freeNodes];
    nrDoF = size(DoF,2);
    
    % Degree of numerical integration
    degreeF = 20;
    
    %% Assembling the parts of stiffness and norm matrix and right hand side
    % Prepare matrices for the local blocks and indices for sparse indexing 
    %for the blocks B, C, D, F and G of the stiffness matrix
    Blocal = zeros(3,2,nrElems);
    IB = zeros(3,2,nrElems);
    JB = zeros(3,2,nrElems);
    Clocal = zeros(3,1,nrElems);
    IC = zeros(3,1,nrElems);
    JC = zeros(3,1,nrElems);
    Dlocal = zeros(3,3,nrElems);
    ID = zeros(3,3,nrElems);
    JD = zeros(3,3,nrElems);
    Flocal = zeros(3,2,nrElems);
    IF = zeros(3,2,nrElems);
    JF = zeros(3,2,nrElems);
    Glocal = zeros(3,3,nrElems);
    IG = zeros(3,3,nrElems);
    JG = zeros(3,3,nrElems);
    % Prepare matrices for the local blocks and indices for sparse indexing 
    %for the inverted blocks Moneinv and Mtwoinv of the norm matrix
    Monelocal = zeros(3,3,nrElems);
    IMone = zeros(3,3,nrElems);
    JMone = zeros(3,3,nrElems);
    Mtwolocal = zeros(3,3,nrElems);
    IMtwo = zeros(3,3,nrElems);
    JMtwo = zeros(3,3,nrElems);
    % Prepare vector for the right hand side data
    bRHS = zeros(3,nrElems);
  
    assembleStart = tic; % Measure the time spent for assembling the system matrix
    
    % Initialize function handles for barycentric coordinates
    [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e);
    % Integrals of the products of f with the
    % local barycentric coordinates on each element
    intFLambdaOne4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda1(Gpts4p)),degreeF,area4e);
    intFLambdaTwo4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda2(Gpts4p)),degreeF,area4e);
    intFLambdaThree4e = integrate(c4n,n4e,...
                        @(n4p,Gpts4p,Gpts4ref)(f(Gpts4p).*...
                        Lambda3(Gpts4p)),degreeF,area4e);
    intFLambda4e = [intFLambdaOne4e, intFLambdaTwo4e, intFLambdaThree4e];
    
    % Compute the local blocks in a loop over all elements in parallel
    parfor curElem = 1:nrElems
        % Compute local blocks B and C
        Blocal(:,:,curElem) = area4e(curElem)*[eye(2);zeros(1,2)];
        Clocal(:,:,curElem) = area4e(curElem)*[0;0;2];
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the barycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Compute local blocks D and F
        Dlocal(:,:,curElem) = -area4e(curElem)*[grads';2/3,2/3,2/3];
        Flocal(:,:,curElem) = area4e(curElem)*grads;
        % Compute the sides of the current element in global enumeration of
        % sides as in n4s
        sides = s4e(curElem,:);
        % Compute the vector containing the lengths of the three sides of
        % the current element, equipped with a sign correspnding to the
        % orientation of the current side: 1 if element-normal equals to
        % edge-normal, -1 if the normals are directed in opposite
        % directions
        signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
        % Compute local block G
        Glocal(:,:,curElem) = -1/2 * [1,0,1;1,1,0;0,1,1].*repmat(signedLength,3,1);
        % Compute a matrix containing the vectors of the three edges
        edges = coords - coords([2,3,1],:);
        % Compute auxiliary matrix H
        H = 1/12 * (ones(3,3)+eye(3));
        % Compute auxiliary matrix E
        E = sum(diag(edges'*edges));
        % Compute the local blocks Mone and Mtwo of the
        % norm matrices
        Monelocal(:,:,curElem) = area4e(curElem) * [eye(2),zeros(2,1);zeros(1,2),(1/36*E+4)];
        Mtwolocal(:,:,curElem) = area4e(curElem) * (H+grads*grads');
        
        % Compute index vectors for the matrices
        % The local index matrices I and J have the same size as the
        % corresponding local matrices, I contains the row number of
        % each entry, J contains the column number of each entry 
        IB(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,2);
        JB(:,:,curElem) = repmat(((2*(curElem-1)+1) : 2*curElem),3,1);
        IC(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,1);
        JC(:,:,curElem) = repmat(((1*(curElem-1)+1) : 1*curElem),3,1);
        ID(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,3);
        JD(:,:,curElem) = repmat(nodes,3,1);        
        IF(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,2);
        JF(:,:,curElem) = repmat(((2*(curElem-1)+1) : 2*curElem),3,1);
        IG(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,3);
        JG(:,:,curElem) = repmat(sides,3,1); 
        IMone(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,3);
        JMone(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem),3,1);
        IMtwo(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,3);
        JMtwo(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem),3,1);
        
        % Compute right hand side vector
        bRHS(:,curElem) = intFLambda4e(curElem,:);
    end

    % Assemble the blocks of the stiffness matrix and of the inverse norm
    % matrix using the index vectors and local matrices computed in the
    % loop
    B = sparse(IB(:),JB(:),Blocal(:));
    C = sparse(IC(:),JC(:),Clocal(:));
    D = sparse(ID(:),JD(:),Dlocal(:));
    F = sparse(IF(:),JF(:),Flocal(:));
    G = sparse(IG(:),JG(:),Glocal(:));
    Mone = sparse(IMone(:),JMone(:),Monelocal(:));
    Mtwo = sparse(IMtwo(:),JMtwo(:),Mtwolocal(:));
    
    % Prepare zero-blocks of stiffness matrix and norm matrix
    O1 = sparse(size(B,1),size(G,2));
    O2 = sparse(size(F,1),size(C,2));
    O3 = sparse(size(F,1),size(D,2));
    O4 = sparse(size(Mone,1),size(Mtwo,2));
    
    % Piece together the stiffness matrix and the norm matrix
    S = [B,C,O1,D;F,O2,G,O3];
    M = [Mone,O4;O4',Mtwo];
    A = [M, S; S', sparse(size(S,2),size(S,2))];
    assembleTime = toc(assembleStart); % Measure the time spent for assembling the system matrix
    
    
    %% Include boundary data    
    % Include Dirichlet boundary conditions
    u = zeros(nrNodes,1);
    valU4Db = u4Db(c4n(nodes4Db,:)); 
    u(nodes4Db) = valU4Db(:,1); 
    bDirichlet = -D*u(:);
    
    % Include Neumann boundary conditions
    t = zeros(nrSides,1);
    % Case with Neumann boundary
    % Use midpoint quadrature rule to compute piecewise constant
    % approximation of Neumann boundary data
    if( nrFreeSides ~= nrSides ) 
        valG = g(mid4s(sides4Nb,:));
        t(sides4Nb,1) = valG(:,1);
    end
    
    % Compute corresponding contributions of bilinear form, scaled
    % w.r.t. the domain
    bNeumann = -G*t(:);

    % Piece together the right hand side vector
    b = [bDirichlet;bRHS(:)+bNeumann; zeros(3*nrElems+nrNodes+nrSides,1)];
    
    solveStart = tic;
    %% Solve system
    x = A(DoF,DoF)\b(DoF);
    solveTime = toc(solveStart); % Measure the time spent for solving
    
    % Pick the parts of the solution
    q = x([1:3:3*nrElems;2:3:3*nrElems;3:3:3*nrElems;]');
    v = x(3*nrElems+[1:3*nrElems]');
    r = x(6*nrElems+[1:2:2*nrElems;2:2:2*nrElems]');
    w = x(8*nrElems+[1:nrElems]');
    t(freeSides,1) = x(9*nrElems+[1:nrFreeSides]');
    u(freeNodes,1) = x(9*nrElems+nrFreeSides+[1:nrFreeNodes]');
    
    % Collect the performance data
    data = [assembleTime, solveTime, condest(A(DoF,DoF))];
end


function [Lambda1, Lambda2, Lambda3] = barycentricCoords(c4n,n4e)
        %% Compute vertices
        P1 = c4n(n4e(:,1),:);
        P2 = c4n(n4e(:,2),:);
        P3 = c4n(n4e(:,3),:);
        
        
        %% Define barycentric coordinates
        % by means of affine transformation Phi x + P3
        % where Phi = [P1-P3,P2-P3] '2x2-matrix' in row notation
        % where [a c; b d] --> [a b c d]
        Phi = [P1-P3,P2-P3];
        % Compute Psi = inverse of Phi
        Psi = repmat(1./(Phi(:,1).*Phi(:,4) - Phi(:,2).*Phi(:,3)),1,4)...
                 .*[Phi(:,4),-Phi(:,2),-Phi(:,3),Phi(:,1)];
        % Barycentric coordinates
        Lambda1 = @(z)(Psi(:,1).*(z(:,1)-P3(:,1)) + Psi(:,3).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda2 = @(z)(Psi(:,2).*(z(:,1)-P3(:,1)) + Psi(:,4).*...
                                                    (z(:,2)-P3(:,2)));
        Lambda3 = @(z)(1 - Lambda1(z) - Lambda2(z));     
end


% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
