function [v, u, nrDoF, data] = solveReducedFormulation(c4n,n4e,n4sDb,n4sNb,f,u4Db,g,alpha,Q)
% Function to compute discrete solution to reduced formulation for the
% Poisson model problem.
% The Poisson model problem seeks u with
%             - Delta u = f         on Omega
%                     u = u4Db      on Dirichlet boundary of Omega
%          nabla u * nu = g         on Neumann boundary of Omega
% given righthand side f, Dirichlet boundary condition u4Db and 
% Neumann boundary condition g on the domain Omega given by c4n and n4e.
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         n4sDb	   nodes of the sides in the Dirichlet boundary
%         n4sNb	   nodes of the sides in the Neumann boundary
%         f	       right hand side function
%         u4Db     Dirichlet boundary condition function
%         g        Neumann boundary condition function
%         alpha    real parameter in reduced formulation
%         Q        local projection in reduced formulation
% Output: v        Crouzeix Raviart approximation v_{CR} of residual
%         u        Courant approximation u_C of u 
%         nrDof    number of degrees of freedom (size of system matrix)
%         data     data array containing time for solve, assembly, postprocess and
%                  condition number of system matrix

    %% Compute mesh quantities
    % Numbers of elements, nodes and sides
    nrElems = size(n4e,1);
    nrNodes = size(c4n,1);
    n4s = computeN4s(n4e);
    nrSides = size(n4s,1);
    
    % Compute areas and sides for elements and nodes
    area4e = computeArea4e(c4n,n4e);
    s4e = computeS4e(n4e);
    s4n = computeS4n(n4e);
   
    % Compute degrees of freedom for nodes
    nodes4Db = unique(n4sDb)';
    freeNodes = setdiff(1:nrNodes,nodes4Db);    
    % Compute degrees of freedom for sides
    sides4Db = unique(s4n(sub2ind(size(s4n),n4sDb(:,1),n4sDb(:,2))))';
    freeSides = setdiff(1:nrSides, sides4Db);
    % Put together all degrees of freedom
    DoF = [freeSides, nrSides + freeNodes];
    nrDoF = size(DoF,2);
    
    %% Assembling the parts of stiffness matrix
    % Prepare local matrices for the blocks B and D of the stiffness matrix
    Blocal = zeros(3,3,nrElems);
    IB = zeros(3,3,nrElems);
    JB = zeros(3,3,nrElems);
    Dlocal = zeros(3,3,nrElems);
    JD = zeros(3,3,nrElems);
    
    assembleStart = tic; % Measure the time spent for assembling the system matrix
    % Compute L^2 products with projections Q, i.e. local blocks C and
    % local right-hand sides
    [Clocal,bRHS] = calculateL2Product(c4n,n4e,f,Q);
    
    % Compute the local blocks for B and D in a loop over elements in parallel
    parfor curElem = 1:nrElems
%     for curElem = 1:nrElems
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the barycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Gradients of CR basis are scaled gradients of barycentric
        % coordinates
        gradsEdges = -2*grads([3 1 2],:);
        % Compute the sides of the current element in global enumeration of
        % sides as in n4s
        sides = s4e(curElem,:);
        % Block B represents energy scalar product of gradients of CR with
        % gradients of CR
        Blocal(:,:,curElem) = area4e(curElem) * gradsEdges * gradsEdges'; 
        % Block D represents energy scalar product of gradients of CR with
        % gradients of barycentric coordinates
        Dlocal(:,:,curElem) = area4e(curElem) * gradsEdges * grads';
        
        % Compute index vectors for the matrices
        % The local index matrices I and J have the same size as the
        % corresponding local matrices, I contains the row number of
        % each entry, J contains the column number of each entry 
        IB(:,:,curElem) = repmat(sides',1,3);    
        JB(:,:,curElem) = repmat(sides,3,1);         
%         IC(:,:,curElem) = repmat(sides',1,3); 
        JD(:,:,curElem) = repmat(nodes,3,1);        
%         ID(:,:,curElem) = repmat(nodes',1,3); 
%         JD(:,:,curElem) = repmat(nodes,3,1); 
    end
    
%     % Index vector for matrices with rows corresponding to Crouzeix-Raviart
%     % functions
%     s4eT = s4e';
%     RowsCR = [s4eT;s4eT;s4eT];
%     % Index vector for matrices with columns corresponding to 
%     % Crouzeix-Raviart functions
%     ColsCR = [s4eT(:),s4eT(:),s4eT(:)]';
%     % Index vector for matrices with columns corresponding to 
%     % Courant functions
%     n4eT = n4e';
%     ColsCourant = [n4eT(:),n4eT(:),n4eT(:)]';
%     
%     % Assemble the blocks of the stiffness matrix using the local matrices 
%     % computed in the loop and index vectors for sparse matrix initiation
%     B = sparse(RowsCR(:),ColsCR(:),Blocal(:));
%     C = sparse(RowsCR(:),ColsCR(:),Clocal(:));
%     D = sparse(RowsCR(:),ColsCourant(:),Dlocal(:));
    B = sparse(IB(:),JB(:),Blocal(:));
    C = sparse(IB(:),JB(:),Clocal(:));
    D = sparse(IB(:),JD(:),Dlocal(:));

    % Piece together the stiffness matrix
    A = [B+alpha*C,D;D', sparse(size(D,2),size(D,2))];
    
    assembleTime = toc(assembleStart); % Measure the time spent for assembling the system matrix
    
    %% Calculate the right hand side vector
    
    % Prepare solution vector
    x = zeros(nrSides + nrNodes, 1);
    % Dirichlet boundary conditions by nodal interpolation
    DbNodes = unique(n4sDb);
    x(nrSides+DbNodes) = u4Db(c4n(DbNodes,:));

    % Neumann boundary conditions
    % Compute length and midpoints of Neumann boundary sides
    length4NbSides = computeLength4s(c4n,n4sNb);
    mid4NbSides = computeMid4s(c4n,n4sNb);
    bNeumann = zeros(nrSides,1);
    sides4Nb = zeros(1,size(n4sNb,1));
    for i = 1:size(n4sNb,1)
        sides4Nb(i) = s4n(n4sNb(i,1),n4sNb(i,2));
    end
    if size(n4sNb,1) > 0
%         % Compute numbers of Neumann boundary sides
%         sides4Nb = unique(s4n(sub2ind(size(s4n),n4sNb(:,1),n4sNb(:,2))))';
        % Use midpoint quadrature rule to compute piecewise constant
        % approximation of Neumann boundary data
        bNeumann(sides4Nb) = length4NbSides .* g(mid4NbSides);
    end
    
    % Piece together the right hand side vector consisting of 
    % right-hand side function (bRHS),
    % Neumann boundary data (bNeumann),
    % Dirichlet boundary data (A*x)
    s4eT = s4e';
    b = [sparse(s4eT(:),ones(3*nrElems,1),bRHS(:)) + bNeumann;zeros(nrNodes,1)] - A*x;
    
    %% Solve the system of equations    
    solveStart = tic;
    x(DoF) = A(DoF,DoF)\b(DoF);
    solveTime = toc(solveStart); % Measure the time spent for solving
    
    % Pick the parts of the solution concerning the 2 components of the
    % solution
    v = x((1:nrSides)');
    u = x(nrSides+1:end);
    
    postprocessStart = tic;
    %% Post-processing in case of parameters that match primal or ultraweak dPG method
    if ((alpha==1 || alpha == 0.5) && strcmp(Q,'identity'))
        length4s = computeLength4s(c4n,n4s);
        normal4e = computeNormal4e(c4n,n4e);
        normal4s = computeNormal4s(c4n,n4s);
        mid4e = computeMid4e(c4n, n4e);
        q = zeros(nrElems,3);
        r = zeros(nrElems,2);
        w = zeros(nrElems,1);
        for curElem = 1:nrElems
            % Compute the nodes of the current element in global enumeration
            nodes = n4e(curElem,:);
            % Compute the coordinates of the nodes of the current element
            coords = c4n(nodes,:);
            % Compute the gradients of the barycentric coordinates of the
            % current element
            grads = [1,1,1;coords'] \ [0,0;eye(2)];
            % Compute the sides of the current element in global enumeration of
            % sides as in n4s
            sides = s4e(curElem,:);
            % Compute the vector containing the lengths of the three sides of
            % the current element, equipped with a sign correspnding to the
            % orientation of the current side: 1 if element-normal equals to
            % edge-normal, -1 if the normals are directed in opposite
            % directions
            signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
            % Compute local block corresponding to trace term <,>
            Tlocal(:,:,curElem) = 1/2 * [1,0,1;1,1,0;0,1,1].*repmat(signedLength,3,1);
            fval = repmat(f(mid4e(curElem,:)),3,1);
            RHS(:,curElem) = area4e(curElem) * grads * grads'*u(nodes)+(Clocal(:,:,curElem)+Dlocal(:,:,curElem)')*v(s4e(curElem,:)) - 1/3 * area4e(curElem) * fval(:);
            IT(:,:,curElem) = repmat(((3*(curElem-1)+1) : 3*curElem)',1,3);
            JT(:,:,curElem) = repmat(sides,3,1);   
            if alpha==0.5
                % Gradients of CR basis are scaled gradients of barycentric
                % coordinates
                gradsEdges = -2*grads([3 1 2],:);
                q(curElem,:) = [- alpha*gradsEdges'*v(s4e(curElem,:)) ;0]';
                r(curElem,:)= alpha*v(s4e(curElem,:))'*gradsEdges  + u(nodes)'*grads;
                w(curElem,:) = 1/3*u(nodes)'*ones(3,1);
            end
        end
        T = sparse(IT(:),JT(:),Tlocal(:));
        t = (T'*T)\(T'*RHS(:));
    end
    postprocessTime = toc(postprocessStart);
    
    % Collect the performance data
    data = [assembleTime, solveTime,postprocessTime, condest(A(DoF,DoF))];
end


% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.