function [p, u, nrDoF, data] = solveWeightedLSFormulation(c4n,n4e,n4sDb,n4sNb,f,u4Db,g,tag_M0,tag_F0)
% Function to compute discrete solution to weighted LS formulation for the
% Poisson model problem.
% The Poisson model problem seeks u with
%             - Delta u = f         on Omega
%                     u = u4Db      on Dirichlet boundary of Omega
%          nabla u * nu = g         on Neumann boundary of Omega
% given righthand side f, Dirichlet boundary condition u4Db and 
% Neumann boundary condition g on the domain Omega given by c4n and n4e.
% Input:  c4n      coordinates for the nodes of the mesh
%         n4e      nodes for the elements of the mesh
%         n4sDb	   nodes of the sides in the Dirichlet boundary
%         n4sNb	   nodes of the sides in the Neumann boundary
%         f	       right hand side function
%         u4Db     Dirichlet boundary condition function
%         g        Neumann boundary condition function
%         tag_M0   tag for choice of M_0 in weighted LS
%         tag_F0   tag for choice of F_0 in weighted LS
% Output: p        Raviart-Thomas approximation p_{RT} of nabla u
%         u        Courant approximation u_C of u 
%         nrDof    number of degrees of freedom (size of system matrix)
%         data     data array containing time for solve, assembly, postprocess and
%                  condition number of system matrix

    %% Compute mesh quantities
    % Numbers of elements, nodes and sides
    nrElems = size(n4e,1);
    nrNodes = size(c4n,1);
    n4s = computeN4s(n4e);
    nrSides = size(n4s,1);

    % Compute areas, mid points, lengths of sides, normals and sides for elements and nodes
    area4e = computeArea4e(c4n,n4e);
    mid4s = computeMid4s(c4n, n4s);
    mid4e = computeMid4e(c4n, n4e);
    length4s = computeLength4s(c4n,n4s);
    normal4e = computeNormal4e(c4n,n4e);
    normal4s = computeNormal4s(c4n,n4s);
    s4e = computeS4e(n4e);
    s4n = computeS4n(n4e);
    
    % Compute degrees of freedom for nodes
    nodes4Db = unique(n4sDb)';
    freeNodes = setdiff(1:nrNodes,nodes4Db);
    
    % Compute degrees of freedom for sides
    if size(n4sNb,1) > 0
        sides4Nb = unique(s4n(sub2ind(size(s4n),n4sNb(:,1),n4sNb(:,2))))';
    else
        sides4Nb = [];
    end
    freeSides = setdiff(1:nrSides, sides4Nb);
    % Put together all degrees of freedom
    DoF = [freeSides,nrSides + freeNodes];
    nrDoF = size(DoF,2);

    %% Assembling the parts of stiffness and norm matrix and right hand side
    % Prepare matrices for the local blocks B, C, and D of the stiffness matrix
    Blocal = zeros(3,3,nrElems);
    IB = zeros(3,3,nrElems);
    JB = zeros(3,3,nrElems);
    Clocal = zeros(3,3,nrElems);
    JC = zeros(3,3,nrElems);
    ID = zeros(3,3,nrElems);
    % Prepare vector for the local right hand side data
    bpRHS = zeros(3,nrElems);
    buRHS = zeros(3,nrElems);

    assembleStart = tic; % Measure the time spent for assembling the system matrix    
    % Compute weight function M_0 (piecewise constant 2x2 SPD matrix)
    M0 = calculateM0(c4n,n4e,tag_M0);
    % Compute right hand side oscillation (piecewise constant 2-vector)
    F0 = calculateF0(c4n,n4e,tag_F0,f);
    % Compute integral mean of f on each element
    degreeF = 20; % degree of numerical integration
    ProjF = integrate(c4n, n4e, @(n4p,Gpts4p,Gpts4ref)f(Gpts4p), ...
                   degreeF, area4e)./area4e;
               
    % Compute the local blocks in a loop over all elements in parallel
    parfor curElem = 1:nrElems
        % Compute the nodes of the current element in global enumeration
        nodes = n4e(curElem,:);
        % Compute the coordinates of the nodes of the current element
        coords = c4n(nodes,:);
        % Compute the gradients of the braycentric coordinates of the
        % current element
        grads = [1,1,1;coords'] \ [0,0;eye(2)];
        % Compute the sides of the current element in global enumeration of
        % sides as in n4s
        sides = s4e(curElem,:);
        % Compute the vector containing the lengths of the three sides of
        % the current element, equipped with a sign correspnding to the
        % orientation of the current side: 1 if element-normal equals to
        % edge-normal, -1 if the normals are directed in opposite
        % directions
        signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
        % Compute matrix representing the piecewise constant parts of the
        % three Raviart-Thomas basis functions of the edges
        Q = 1/(2*area4e(curElem))*repmat(signedLength',1,2).* (repmat(mid4e(curElem,:),3,1)-coords([3,1,2],:));
        
        % Block B represents L^2 scalar product of divergence of RT basis
        % (contained in signedLength) plus L^2 scalar product of piecewise 
        % constant part of RT basis (contained in Q) weighted with M0
        Blocal(:,:,curElem) = (1/area4e(curElem))*(signedLength'*signedLength) + area4e(curElem)*Q*inv(M0(:,:,curElem))*Q';
        % Block C represents L^2 scalar product of piecewise constant part 
        % of RT basis (contained in Q) with gradients of barycentric 
        % coordinates weighted with M0
        Clocal(:,:,curElem) = -area4e(curElem)*Q*inv(M0(:,:,curElem))*grads';
        % Block D represents energy scalar product of gradients of barycentric 
        % coordinates with gradients of barycentric coordinates weighted with M0
        Dlocal(:,:,curElem) = area4e(curElem)*grads*inv(M0(:,:,curElem))*grads';
        
        % Compute index vectors for the matrices
        % The local index matrices I and J have the same size as the
        % corresponding local matrices, I contains the row number of
        % each entry, J contains the column number of each entry 
        IB(:,:,curElem) = repmat(sides',1,3);    
        JB(:,:,curElem) = repmat(sides,3,1);         
%         IC(:,:,curElem) = repmat(sides',1,3); 
        JC(:,:,curElem) = repmat(nodes,3,1);        
        ID(:,:,curElem) = repmat(nodes',1,3); 
%         JD(:,:,curElem) = repmat(nodes,3,1); 
        
        
        % Compute right hand side vector
        % Compute part of right hand side concerning variable p
        bpRHS(:,curElem) = -ProjF(curElem,:) *signedLength' - area4e(curElem)*Q*inv(M0(:,:,curElem))*F0(curElem,:)';
        % Compute part of right hand side concerning variable u
        buRHS(:,curElem) = area4e(curElem)*grads*inv(M0(:,:,curElem))*F0(curElem,:)';
    end

%     % Index vector for matrices with rows corresponding to Raviart-Thomas
%     % functions
%     s4eT = s4e';
%     RowsRT = [s4eT;s4eT;s4eT];
%     % Index vector for matrices with columns corresponding to 
%     % Raviart-Thomas functions
%     ColsRT = [s4eT(:),s4eT(:),s4eT(:)]';
%     % Index vector for matrices with columns corresponding to 
%     % Courant functions
%     n4eT = n4e';
%     RowsCourant = [n4eT;n4eT;n4eT];
%     ColsCourant = [n4eT(:),n4eT(:),n4eT(:)]';
    
    % Assemble the blocks of the stiffness matrix using the local matrices 
    % computed in the loop and index vectors for sparse matrix initiation
%     B = sparse(RowsRT(:),ColsRT(:),Blocal(:));
%     C = sparse(RowsRT(:),ColsCourant(:),Clocal(:));
%     D = sparse(RowsCourant(:),ColsCourant(:),Dlocal(:));
    B = sparse(IB(:),JB(:),Blocal(:));
    C = sparse(IB(:),JC(:),Clocal(:));
    D = sparse(ID(:),JC(:),Dlocal(:));
    
    % Piece together the stiffness matrix
    A = [B,C;C',D];
    assembleTime = toc(assembleStart); % Measure the time spent for assembling the system matrix    
    
    %% Calculate the right hand side vector
    
    % Prepare solution vector
    x = zeros(nrSides + nrNodes, 1);
    % Dirichlet boundary conditions by nodal interpolation
    DbNodes = unique(n4sDb);
    x(nrSides+DbNodes) = u4Db(c4n(DbNodes,:));
    
    % Neumann boundary conditions
    % Use midpoint quadrature rule to compute piecewise constant
    % approximation of Neumann boundary data
    if size(n4sNb,1) > 0
        x(sides4Nb) = g(mid4s(sides4Nb,:));
    end

    % Piece together the right hand side vector consisting of 
    % right-hand side part concerning variable p (bpRHS),
    % right-hand side part concerning variable u (buRHS),
    % Dirichlet and Neumann boundary data (A*x
    s4eT = s4e';
    n4eT = n4e';
    b = [sparse(s4eT(:),ones(3*nrElems,1),bpRHS(:));sparse(n4eT,ones(3*nrElems,1),buRHS(:))] - A*x;
    
    %% Solve the system of equations
    solveStart = tic;
    x(DoF) =  A(DoF,DoF)\b(DoF);
    solveTime = toc(solveStart); % Measure the time spent for solving
    
    % Pick the parts of the solution concerning the 2 parts of the
    % solution
    p = x((1:nrSides)');
    u = x(nrSides+1:end);
    
    postprocessStart = tic;
    %% Post-processing in case of parameters that match primal or ultraweak dPG method
    if ((strcmp(tag_M0,'identitySZero') || strcmp(tag_M0,'twoidentitySZero')) && strcmp(tag_F0,'H1'))
        length4s = computeLength4s(c4n,n4s);
        normal4e = computeNormal4e(c4n,n4e);
        normal4s = computeNormal4s(c4n,n4s);
        mid4e = computeMid4e(c4n, n4e);
        vLocal = zeros(3,nrElems);
        q = zeros(nrElems,3);
        r = zeros(nrElems,2);
        w = zeros(nrElems,1);
        parfor curElem = 1:nrElems
            % Compute the nodes of the current element in global enumeration
            nodes = n4e(curElem,:);
            % Compute the coordinates of the nodes of the current element
            coords = c4n(nodes,:);
            % Compute the gradients of the barycentric coordinates of the
            % current element
            grads = [1,1,1;coords'] \ [0,0;eye(2)];
            % Compute the sides of the current element in global enumeration of
            % sides as in n4s
            sides = s4e(curElem,:);
            % Compute the vector containing the lengths of the three sides of
            % the current element, equipped with a sign correspnding to the
            % orientation of the current side: 1 if element-normal equals to
            % edge-normal, -1 if the normals are directed in opposite
            % directions
            signedLength = sum(normal4s(sides,:)'.*normal4e(:,:,curElem)') .* length4s(sides,:)';
            % Compute matrix representing the piecewise constant parts of the
            % three Raviart-Thomas basis functions of the edges
            Q = 1/(2*area4e(curElem))*repmat(signedLength',1,2).* (repmat(mid4e(curElem,:),3,1)-coords([3,1,2],:));

            Projv = ProjF(curElem,:)+ 1/area4e(curElem)*signedLength*p(s4e(curElem,:));
            nablav = -inv(M0(:,:,curElem))*(grads'*u(nodes)-Q'*p(s4e(curElem,:))-F0(curElem,:)');
            vLocal(:,curElem) = Projv*ones(3,1) + (coords-repmat(mid4e(curElem,:),3,1))*nablav;   
            if strcmp(tag_M0,'twoidentitySZero')
                q(curElem,:) = [(inv(M0(:,:,curElem))*(grads'*u(nodes)-Q'*p(s4e(curElem,:))-F0(curElem,:)'))',0]';
                r(curElem,:)= grads'*u(nodes)-inv(M0(:,:,curElem))*(grads'*u(nodes)-Q'*p(s4e(curElem,:))-F0(curElem,:)');
                w(curElem,:) = 1/3*ones(1,3)*u(nodes);
            end
        end
        v = vLocal(:);
    end
    postprocessTime = toc(postprocessStart);
    
    % Collect the performance data
    data = [assembleTime, solveTime,postprocessTime, condest(A(DoF,DoF))];
end



% Copyright 2018 Friederike Hellwig
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
